//Stegano BMP - An MLSoft program
//The program will hide the file in parameter 3 in the bitmap specified in parameter 1,
//storing the result in parameter 2.
//The file can be retrieved using Stegano BMP Retrieve

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"
#include "hide.h"
#include "BMPhide.h"

using namespace std;

int steganoBmpHide(int argc, char **argv, hideArgs *cachedArgs)
{
	//for() counter variable
	int i;

	//Check if enough parameters are given
	cout << "You are now running the Stegano BMP function of MLSS.\n" << endl;
	if(argc < 3)
	{
		showHideHelp();
		return 0;
	}

	hideArgs args;
	if(cachedArgs)
		args = *cachedArgs;
	else
		args = getHideArgs(argc, argv);
	if(args.error)
		return 0;

	//File names
	string hostfile = args.hostfile, outputfile = args.outputfile, guestfile = args.guestfile;
	bool scramble = args.scramble;
	encryptClass *encryption = args.encryption;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = args.bitsperbyte;
	void (*transferByte)(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption) = args.transferByte;

	//hostfile bytes needed for each guestfile byte
	int bytesperbyte = 8 / bitsperbyte;

	//Check if all necessary parameters are given
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}
	if(guestfile.empty())
	{
		cout << "Missing the non-optional option -g!" << endl;
		return 0;
	}
	if(outputfile.empty())
		outputfile = "output.bmp";

	//retrieve file size information
	int inputSize = getFileSize(hostfile);
	int hideSize = getFileSize(guestfile);

	//Open existing BMP
	ifstream bmpIn(hostfile.c_str(), ios::binary);
	if(bmpIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Open existing file that should be hidden
	ifstream hideIn(guestfile.c_str(), ios::binary);
	if(hideIn.fail())
	{
		cout << "Failed to open " << guestfile << endl;
		return 0;
	}

	//Check header of BMP file
	const int HEADERSIZE = 18;
	char perfectHeader[HEADERSIZE] = { 'B', 'M', 0, 0, 0, 0, 0, 0, 0, 0,
		54, 0, 0, 0, 40, 0, 0, 0 };
	//Copy input size
	for(i = 0; i < 4; ++i)
	{
		perfectHeader[i + 2] = (reinterpret_cast<char*>(&inputSize))[i];
	}
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = static_cast<char>(bmpIn.get())) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<unsigned int>(temp) << ") differs from " << static_cast<unsigned int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned int width, height;
	unsigned short planes, colorDepth;
	bmpIn.read(reinterpret_cast<char *>(&width), sizeof(width));
	bmpIn.read(reinterpret_cast<char *>(&height), sizeof(height));
	bmpIn.read(reinterpret_cast<char *>(&planes), sizeof(planes));
	if(planes != 1)
	{
		cout << "Bitmaps with more than 1 plane are not supported." << endl;
		return 0;
	}
	bmpIn.read(reinterpret_cast<char *>(&colorDepth), sizeof(colorDepth));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	int compression;
	bmpIn.read(reinterpret_cast<char *>(&compression), sizeof(compression));
	if(compression != 0)
	{
		cout << "Compressed bitmaps are not supported!" << endl;
		return 0;
	}
	//Read data size
	int dataSize;
	bmpIn.read(reinterpret_cast<char *>(&dataSize), sizeof(dataSize));

	//Vertical and horizontal resolution (dpi)
	int vdpi, hdpi;
	bmpIn.read(reinterpret_cast<char *>(&vdpi), sizeof(vdpi));
	bmpIn.read(reinterpret_cast<char *>(&hdpi), sizeof(hdpi));
	//Number of colors used can be ignored
	int numberOfColors;
	bmpIn.read(reinterpret_cast<char *>(&numberOfColors), sizeof(numberOfColors));
	//Important colors can be ignored
	int importantColors;
	bmpIn.read(reinterpret_cast<char *>(&importantColors), sizeof(importantColors));

	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}

		if(bitsperbyte == 2)
		{
			transferByte = transfer16Byte2;
			bytesperbyte = 4;
		}
		else if(bitsperbyte == 1)
		{
			transferByte = transfer16Byte1;
			bytesperbyte = 6;
		}
	}

	//We need to save the filename (stringlength + 1), the filesize (4) and the actual data (outputSize)
	int neededBytes = bytesperbyte * (static_cast<int>(guestfile.length()) + 1 + 4 + hideSize);
	int availibleBytes = width * (colorDepth >> 3);
	//Check if additional zeros were appended to each line
	while((availibleBytes & 3) > 0) ++availibleBytes;
	availibleBytes *= height;
	//Check if image is big enough
	if(neededBytes > availibleBytes)
	{
		cout << "Hostfile is too small by " << neededBytes - availibleBytes << " bytes!" << endl;
		return 0;
	}

	cout << "Hostfile was checked and approved.\nBeginning to hide " << guestfile << " with " << hideSize << " bytes." << endl;

	//Open output BMP file
	ofstream bmpOut(outputfile.c_str(), ios::binary);

	//See if any unused bytes are placed behind the bitmap data and
	//adjust the header accordingly
	if(54 + availibleBytes != inputSize)
	{
		inputSize = 54 + availibleBytes;
		//Copy input size
		for(i = 0; i < 4; ++i)
		{
			perfectHeader[i + 2] = (reinterpret_cast<char*>(&inputSize))[i];
		}
	}
	//Write header data to output TGA file
	for(i = 0; i < HEADERSIZE; ++i)
		bmpOut.write(&perfectHeader[i], 1);
	bmpOut.write(reinterpret_cast<char *>(&width), sizeof(width));
	bmpOut.write(reinterpret_cast<char *>(&height), sizeof(height));
	bmpOut.write(reinterpret_cast<char *>(&planes), sizeof(planes));
	bmpOut.write(reinterpret_cast<char *>(&colorDepth), sizeof(colorDepth));
	bmpOut.write(reinterpret_cast<char *>(&compression), sizeof(compression));
	bmpOut.write(reinterpret_cast<char *>(&dataSize), sizeof(dataSize));
	bmpOut.write(reinterpret_cast<char *>(&vdpi), sizeof(vdpi));
	bmpOut.write(reinterpret_cast<char *>(&hdpi), sizeof(hdpi));
	bmpOut.write(reinterpret_cast<char *>(&numberOfColors), sizeof(numberOfColors));
	bmpOut.write(reinterpret_cast<char *>(&importantColors), sizeof(importantColors));

	//Perform actual hiding operation
	hideData(bmpIn, bmpOut, hideIn, guestfile, scramble, encryption, bytesperbyte, hideSize,
		transferByte, availibleBytes);
	if(encryption)
		delete encryption;
	cout << "File was successfully hidden." << endl;
	return 0;
}
