//Stegano TGA Retrieve - A Polecatsoft program
//The program will retrieve a file which was hidden using Stegano TGA,
//in the file specified in parameter 1.

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"
#include "retrieve.h"
#include "BMPretrieve.h"

using namespace std;

int steganoBmpRetrieve(int argc, char **argv, retrieveArgs *cachedArgs)
{
	//for() counter variable
	int i;

	//Check if enough parameters are given
	cout << "You are now running the Stegano BMP Retrieve function of MLSS.\n" << endl;
	if(argc < 3)
	{
		showRetrieveHelp();
		return 0;
	}

	retrieveArgs args;
	if(cachedArgs)
		args = *cachedArgs;
	else
		args = getRetrieveArgs(argc, argv);
	if(args.error)
		return 0;

	//File names
	string hostfile = args.hostfile;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = args.bitsperbyte;
	encryptClass *encryption = args.encryption;
	unsigned char (*retrieveByte)(ifstream &input, encryptClass *encryption) = args.retrieveByte;

	//hostfile
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}
	int inputSize = getFileSize(hostfile);

	//Open existing BMP
	ifstream bmpIn(hostfile.c_str(), ios::binary);
	if(bmpIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Check header of BMP file
	const int HEADERSIZE = 18;
	char perfectHeader[HEADERSIZE] = { 'B', 'M', 0, 0, 0, 0, 0, 0, 0, 0,
		54, 0, 0, 0, 40, 0, 0, 0 };
	//Copy input size
	for(i = 0; i < 4; ++i)
	{
		perfectHeader[i + 2] = (reinterpret_cast<char*>(&inputSize))[i];
	}
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = static_cast<char>(bmpIn.get())) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<int>(temp) << ") differs from " << static_cast<int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned int width, height;
	unsigned short planes, colorDepth;
	bmpIn.read(reinterpret_cast<char *>(&width), sizeof(width));
	bmpIn.read(reinterpret_cast<char *>(&height), sizeof(height));
	bmpIn.read(reinterpret_cast<char *>(&planes), sizeof(planes));
	if(planes != 1)
	{
		cout << "Bitmaps with more than 1 plane are not supported." << endl;
		cout << planes << endl;
		return 0;
	}
	bmpIn.read(reinterpret_cast<char *>(&colorDepth), sizeof(colorDepth));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	int compression;
	bmpIn.read(reinterpret_cast<char *>(&compression), sizeof(compression));
	if(compression != 0)
	{
		cout << "Compressed bitmaps are not supported!" << endl;
		return 0;
	}
	//Read data size
	int dataSize;
	bmpIn.read(reinterpret_cast<char *>(&dataSize), sizeof(dataSize));

	//Vertical and horizontal resolution (dpi)
	int vdpi, hdpi;
	bmpIn.read(reinterpret_cast<char *>(&vdpi), sizeof(vdpi));
	bmpIn.read(reinterpret_cast<char *>(&hdpi), sizeof(hdpi));
	//Number of colors used can be ignored
	int numberOfColors;
	bmpIn.read(reinterpret_cast<char *>(&numberOfColors), sizeof(numberOfColors));
	//Important colors can be ignored
	int importantColors;
	bmpIn.read(reinterpret_cast<char *>(&importantColors), sizeof(importantColors));

	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}

		if(bitsperbyte == 2)
			retrieveByte = retrieve16Byte2;
		else if(bitsperbyte == 1)
			retrieveByte = retrieve16Byte1;
	}

	cout << "Input BMP file was checked and approved.\nBeginning to retrieve data." << endl;

	//Retrieve filename
	string filename;
	char readByte;
	while((readByte = retrieveByte(bmpIn, encryption)) != 0)
		filename += readByte;
	//Retrieve filesize
	int fileSize;
	for(i = 0; i < 4; ++i)
		(reinterpret_cast<char *>(&fileSize))[i] = retrieveByte(bmpIn, encryption);
	cout << "Creating the file " << filename << " with " << fileSize << " bytes." << endl;
	cout << endl;

	//Retrieve file content
	ofstream retrovenFile(filename.c_str(), ios::binary);
	if(retrovenFile.fail())
	{
		cout << "Failed to open " << filename << ", maybe no data is actually hidden in the image,\n"
			"or you supplied a wrong value for -b!" << endl;
		return 0;
	}
	readByte = retrieveByte(bmpIn, encryption);
	for(i = 0; i < fileSize; ++i, readByte = retrieveByte(bmpIn, encryption))
		retrovenFile.write(&readByte, 1);
	retrovenFile.close();
	bmpIn.close();

	if(encryption)
		delete encryption;
	cout << "File was successfully retrieved." << endl;
	return 0;
}
