//Stegano TGA Retrieve - A Polecatsoft program
//The program will retrieve a file which was hidden using Stegano TGA,
//in the file specified in parameter 1.

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"
#include "retrieve.h"
#include "TGAretrieve.h"

using namespace std;

int steganoTgaRetrieve(int argc, char **argv, retrieveArgs *cachedArgs)
{
	cout << "You are now running the Stegano TGA Retrieve function of MLSS.\n" << endl;
	//Check if enough parameters are given
	if(argc < 3)
	{
		showRetrieveHelp();
		return 0;
	}
	retrieveArgs args;
	if(cachedArgs)
		args = *cachedArgs;
	else
		args = getRetrieveArgs(argc, argv);
	if(args.error)
		return 0;
	//for() counter variable
	int i;

	//File names
	string hostfile = args.hostfile;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = args.bitsperbyte;
	encryptClass *encryption = args.encryption;
	unsigned char (*retrieveByte)(ifstream &input, encryptClass *encryption) = args.retrieveByte;

	//hostfile
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}

	//Open existing TGA
	ifstream tgaIn(hostfile.c_str(), ios::binary);
	if(tgaIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Check header of TGA file
	const int HEADERSIZE = 12;
	char perfectHeader[HEADERSIZE] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = static_cast<char>(tgaIn.get())) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<unsigned int>(temp) << ") differs from " << static_cast<unsigned int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned short width, height;
	char colorDepth;
	tgaIn.read(reinterpret_cast<char *>(&width), sizeof(short));
	tgaIn.read(reinterpret_cast<char *>(&height), sizeof(short));
	tgaIn.read(&colorDepth, sizeof(char));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	//Descriptor byte
	unsigned char dByte;
	tgaIn.read(reinterpret_cast<char *>(&dByte), sizeof(char));
	bool error = false;
	switch(colorDepth)
	{
	case 16:
		if((dByte & 0x0F) != 0x01)
			error = true;
		break;
	case 24:
		if((dByte & 0x0F) != 0x00)
			error = true;
		break;
	case 32:
		if((dByte & 0x0F) != 0x08)
			error = true;
		break;
	default:
		error = true;
		break;
	}
	if(dByte & 0xF0)
		error = true;
	if(error)
	{
		cout << "Error in image descriptor byte!\nHostfile is not suitable!" << endl;
		return 0;
	}
 
	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}
		if(bitsperbyte == 2)
			retrieveByte = retrieve16Byte2;
		else if(bitsperbyte == 1)
			retrieveByte = retrieve16Byte1;
	}

	cout << "Input TGA file was checked and approved.\nBeginning to retrieve data." << endl;

	//Retrieve all data
	if(retrieveData(tgaIn, encryption, retrieveByte))
		cout << "File was successfully retrieved." << endl;
	tgaIn.close();
	if(encryption)
		delete encryption;
	return 0;
}
