//Common Stegano functions

#include <fstream>
#include <iostream>

#include "main.h"
#include "XorEncryption.h"
#include "hide.h"
#include "BMPhide.h"
#include "TGAhide.h"

using namespace std;

int steganoHide(int argc, char **argv)
{
	if(argc < 3)
	{
		showHideHelp();
		return 0;
	}
	cout << "Determining file type to pass call on to subroutine..." << endl;
	hideArgs args = getHideArgs(argc, argv);
	if(args.error)
		return 0;
	//determine filetype
	string ending = getSuffix(args.hostfile);
	if(ending == "tga" || ending == "TGA")
	{
		cout << "Determined file type to be TGA.\n" << endl;
		return steganoTgaHide(argc, argv, &args);
	}
	if(ending == "bmp" || ending == "BMP")
	{
		cout << "Determined file type to be BMP.\n" << endl;
		return steganoBmpHide(argc, argv, &args);
	}
	cout << "The file suffix " << ending << " is not associated with a hide function.\n"
		"Please set the filetype manually by directly calling\n"
		"tgahide (for TGA files) or bmphide (for BMP files)." << endl;
	return 0;
}

void hideData(ifstream &in, ofstream &out, ifstream &hideIn, string guestfile,
			  bool scramble, encryptClass *encryption, int bytesperbyte, int hideSize,
			  void (*transferByte)(ifstream &input, ofstream &output,
					unsigned char writeByte, encryptClass *encryption),
			  int availibleBytes)
{
	//Perform actual hiding operation
	int filePos = 0, i;
	//Hide filename as 0 terminated string
	for(i = 0; i < int(guestfile.length()); ++i, filePos += bytesperbyte)
		transferByte(in, out, static_cast<unsigned char>(guestfile[i]), encryption);
	transferByte(in, out, 0, encryption);
	filePos += bytesperbyte;
	//Hide hidden file size
	for(i = 0; i < 4; ++i, filePos += bytesperbyte)
		transferByte(in, out, (reinterpret_cast<char *>(&hideSize))[i], encryption);
	//Hide the file
	for(i = 0; i < hideSize; ++i, filePos += bytesperbyte)
	{
		char tmp;
		hideIn.read(&tmp, 1);
		transferByte(in, out, static_cast<unsigned char>(tmp), encryption);
	}
	hideIn.close();
	if(encryption)
		delete encryption;
	//Transfer the rest of the file
	if(filePos < availibleBytes)
	{
		cout << "Beginning to copy " << availibleBytes - filePos << " unused bytes." << endl;
		if(scramble)
		{
			cout << "Scrambling up the unused bytes due to -s,\nthis might be very slow." << endl;
			for(i = 0; i < availibleBytes - filePos; ++i)
				transferByte(in, out, static_cast<unsigned char>(rand() & 0xFF), 0);
		}
		else
		{
			char *tmp = new char[availibleBytes - filePos];
			in.read(tmp, availibleBytes - filePos);
			out.write(tmp, availibleBytes - filePos);
			delete [] tmp;
		}
	}
}

hideArgs::hideArgs()
{
	error = false;
	scramble = false;
	encryption = 0;
	transferByte = transferByte1;
	bitsperbyte = 1;
}

hideArgs getHideArgs(int argc, char **argv)
{
	hideArgs args;
	for(int i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-h"))
		{
			if(++i < argc)
			{
				args.hostfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-o"))
		{
			if(++i < argc)
			{
				args.outputfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-g"))
		{
			if(++i < argc)
			{
				args.guestfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-b"))
		{
			if(++i < argc)
			{
				args.bitsperbyte = str2int(argv[i]);
				if(args.bitsperbyte == 1)
					args.transferByte = transferByte1;
				else if(args.bitsperbyte == 2)
					args.transferByte = transferByte2;
				else if(args.bitsperbyte == 4)
					args.transferByte = transferByte4;
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-p"))
		{
			if(++i < argc)
			{
				args.encryption = new encryptClass(argv[i]);
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-s"))
		{
			args.scramble = true;
			continue;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		args.error = true;
		return args;
	}
	return args;
}

//For 16 bit images:
//This function hides the writeByte byte in 6 bytes of input, storing the result in output
void transfer16Byte1(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	unsigned short readData;
	//Hide the first 6 bits from writeByte
	for(int i = 0; i < 6; i += 3)
	{
		input.read(reinterpret_cast<char*>(&readData), 2);
		//This instruction performs all the code below in one step
		readData = (readData & 0xFBDE) | ((writeByte & (1 << i)) << (10 - i))
			                           | ((writeByte & (1 << (i + 1))) << (4 - i)) // == 5 - (i + 1)
									   | ((writeByte >> (i + 2)) & 0x0001);
									   //| ((writeByte & (1 << (i + 2))) >> (i + 2));
#if 0
		//Delete 3 bits from the read data
		readData &= 0xFBDE; //  A RRRRR GGGGG BBBBB
		                    //& 1 11110 11110 11110
		//Get and set first bit
		unsigned char curBit = (writeByte & (1 << i)) >> i;
		readData |= curBit << 10;
		//Get and set second bit
		curBit = (writeByte & (1 << (i + 1))) >> (i + 1);
		readData |= curBit << 5;
		//Get and set third bit
		curBit = (writeByte & (1 << (i + 2))) >> (i + 2);
		readData |= curBit;
#endif
		output.write(reinterpret_cast<char*>(&readData), 2);
	}
	//Hide the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);
	//Delete 2 bits from the read data and set the 2 bits from the writeByte
	//  A RRRRR GGGGG BBBBB
	//& 1 11110 11110 11111
	readData = (readData & 0xFBDF) | ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
	output.write(reinterpret_cast<char*>(&readData), 2);
}

//This function hides the writeByte byte in 4 bytes of input, storing the result in output
void transfer16Byte2(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	//Hide the first 6 bits from writeByte
	unsigned short readData;
	input.read(reinterpret_cast<char*>(&readData), 2);
	//This instruction performs all the code below in one step
	readData = (readData & 0xF39C) | ((writeByte & 0x03) << 10)
		                           | ((writeByte & 0x0C) << 3)
		                           | ((writeByte & 0x30) >> 4);
#if 0
	//Delete 6 bits from the read data
	readData &= 0xF39C; //  A RRRRR GGGGG BBBBB
	                    //& 1 11100 11100 11100
	//Get and set first two bits
	unsigned char curBit = writeByte & 0x03;
	readData |= curBit << 10;
	//Get and set second two bits
	curBit = (writeByte & 0x0C) >> 2;
	readData |= curBit << 5;
	//Get and set third two bits
	curBit = (writeByte & 0x30) >> 4;
	readData |= curBit;
#endif
	output.write(reinterpret_cast<char*>(&readData), 2);

	//Hide the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);
	readData = (readData & 0xFBDF) | ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
#if 0
	//Delete 2 bits from the read data
	readData &= 0xFBDF; //  A RRRRR GGGGG BBBBB
	                    //& 1 11110 11110 11111
	//Set the 2 bits from the writeByte
	readData |= ((writeByte & 0x40) << 4) | ((writeByte & 0x80) >> 2);
#endif
	output.write(reinterpret_cast<char*>(&readData), 2);
}

//For 24 and 32 bit images:
//This function hides the writeByte byte in 8 bytes of input, storing the result in output
void transferByte1(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	unsigned char readByte;
	for(int i = 0; i < 8; ++i)
	{
		//Get value of active bit
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last bit to new value
		readByte = (readByte & 0xFE) | ((writeByte >> i) & 1);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

//This function hides the writeByte byte in 4 bytes of input, storing the result in output
void transferByte2(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	for(int i = 0; i < 8; i += 2)
	{
		unsigned char readByte;
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last 2 bits to new value
		readByte = (readByte & 0xFC) | ((writeByte >> i) & 0x03);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

//This function hides the writeByte byte in 2 bytes of input, storing the result in output
void transferByte4(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption)
{
	if(encryption)
		encryption->exe(writeByte);
	for(int i = 0; i < 8; i += 4)
	{
		unsigned char readByte;
		input.read(reinterpret_cast<char *>(&readByte), 1);
		//Set last 4 bits to new value
		readByte = (readByte & 0xF0) | ((writeByte >> i) & 0x0F);
		output.write(reinterpret_cast<char *>(&readByte), 1);
	}
}

void showHideHelp()
{
	cout << "Possible options (explanations are in brackets):\n"
		<< "-h hostfile (image file which will serve as host)\n"
		<< "-o outputfile (image file that will be created; default: output.tga\nor output.bmp, respectively)\n"
		<< "-g guestfile (file that will be hidden)\n"
		<< "-p password (a password that will be used to encrypt the guestfile)\n"
		<< "-s (scramble unused parts of the file using random data)\n"
		<< "-b bitsperbyte (defines how many bits of each byte of the hostfile will be used to hide data. Possible values are 1, 2 & 4 (only 1 & 2 for 16 bit images);\ndefault: 1)\n";

	cout << "Non-optional options: -h -g" << endl;
}
