//Master Lenman's Steganography Suite - MLSS
//Entry point for the application.
//The main function decides which function to call.

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"
#include "hide.h"
#include "retrieve.h"
#include "TGAhide.h"
#include "TGAretrieve.h"
#include "BMPhide.h"
#include "BMPretrieve.h"
#include "mltc.h"

using namespace std;

const char
BIG_VERSION_NUMBER = '0',
SMALL_VERSION_NUMBER = '6';

int main(int argc, char **argv)
{
	cout << "Welcome to Master Lenman's Steganography Suite!\n"
		<< "You are running MLSS " << BIG_VERSION_NUMBER << "." << SMALL_VERSION_NUMBER << "\n"
		<< "<lenman@users.sourceforge.net>\n" << endl;
	if(argc < 2)
	{
		cout << "Usage: function <options>\n\n"
			<< "Possible functions:\n"
			<< "hide - Hides a file in a TGA or BMP image\n"
			<< "retrieve - Retrieves a hidden file from a TGA or BMP image\n"
			<< "mlxe - Master Lenman's XOR Encryption 2.0\n"
			<< "tcompress - Master Lenman's Text Compression\n"
			<< "tdecompress - Master Lenman's Text Decompression\n"
			<< "\nThe options for a specific function can be found out by passing\n"
			<< "the function name as the only parameter." << endl;
		return 0;
	}
	else
	{
		string temp(argv[1]);
		if(temp == string("hide"))
		{
			return steganoHide(argc, argv);
		}
		if(temp == string("retrieve"))
		{
			return steganoRetrieve(argc, argv);
		}
		if(temp == string("tgahide"))
		{
			return steganoTgaHide(argc, argv);
		}
		if(temp == string("tgaretrieve"))
		{
			return steganoTgaRetrieve(argc, argv);
		}
		if(temp == string("bmphide"))
		{
			return steganoBmpHide(argc, argv);
		}
		if(temp == string("bmpretrieve"))
		{
			return steganoBmpRetrieve(argc, argv);
		}
		if(temp == string("mlxe"))
		{
			return XorEncryption(argc, argv);
		}
		if(temp == string("tcompress"))
		{
			return mltcompress(argc, argv);
		}
		if(temp == string("tdecompress"))
		{
			return mltdecompress(argc, argv);
		}
		cout << "Unknown function: " << temp << endl;
	}

	return 0;
}

//Returns a string version of an integer
string int2str(int i)
{
	ostringstream oss; // create a stream
	oss << i; // insert value to stream
	return oss.str(); // extract value and return it
}

//Returns the file suffix from any path
string getSuffix(string path)
{
	size_t index = path.find_last_of('.');
	string ending;
	if(index != string::npos)
		ending = path.c_str() + int(index + 1);
	return ending;
}

//Returns an integer version of a string
int str2int(string strng)
{
	istringstream oss(strng);
	int r;
	if(!(oss >> r))
		r = -1;
	return r;
}

//Returns filesize in byte
unsigned int getFileSize(string filename)
{
	ifstream file(filename.c_str());
	if(file.fail())
		return 0;
	file.seekg(0, ios::end);
	return file.tellg();
}

//Shows a byte as binary code in std::cout
void byteAsBin(unsigned char byte, ostream *target)
{
	*target << "0y";
	for(int i = 0; i < 8; ++i)
	{
		*target << int2str((byte >> (7 - i)) & 1);
	}
}
