#include <fstream>
#include <iostream>

#include "main.h"
#include "XorEncryption.h"
#include "retrieve.h"
#include "TGAretrieve.h"
#include "BMPretrieve.h"

using namespace std;

int steganoRetrieve(int argc, char **argv)
{
	if(argc < 3)
	{
		showRetrieveHelp();
		return 0;
	}
	cout << "Determining file type to pass call on to subroutine..." << endl;
	retrieveArgs args = getRetrieveArgs(argc, argv);
	if(args.error)
		return 0;
	//determine filetype
	string ending = getSuffix(args.hostfile);
	if(ending == "tga" || ending == "TGA")
	{
		cout << "Determined file type to be TGA.\n" << endl;
		return steganoTgaRetrieve(argc, argv, &args);
	}
	if(ending == "bmp" || ending == "BMP")
	{
		cout << "Determined file type to be BMP.\n" << endl;
		return steganoBmpRetrieve(argc, argv, &args);
	}
	cout << "The file suffix " << ending << " is not associated with a retrieve function.\n"
		"Please set the filetype manually by directly calling\n"
		"tgaretrieve (for TGA files) or bmpretrieve (for BMP files)." << endl;
	return 0;
}

bool retrieveData(std::ifstream &in, encryptClass *encryption, unsigned char (*retrieveByte)(ifstream &input, encryptClass *encryption))
{
	//Retrieve filename
	string filename;
	char readByte;
	while((readByte = retrieveByte(in, encryption)) != 0)
		filename += readByte;
	//Retrieve filesize
	int fileSize;
	for(int i = 0; i < 4; ++i)
		(reinterpret_cast<char *>(&fileSize))[i] = retrieveByte(in, encryption);
	cout << "Creating the file " << filename << " with " << fileSize << " bytes." << endl;
	cout << endl;

	//Retrieve file content
	ofstream retrovenFile(filename.c_str(), ios::binary);
	if(retrovenFile.fail())
	{
		cout << "Failed to open " << filename << ", maybe no data is actually hidden in the image,\n"
			"or you supplied a wrong value for -b!" << endl;
		return false;
	}
	readByte = retrieveByte(in, encryption);
	for(int i = 0; i < fileSize; ++i, readByte = retrieveByte(in, encryption))
		retrovenFile.write(&readByte, 1);
	retrovenFile.close();

	return true;
}

retrieveArgs::retrieveArgs()
{
	error = false;
	retrieveByte = retrieveByte1;
	encryption = 0;
	bitsperbyte = 1;
}

retrieveArgs getRetrieveArgs(int argc, char **argv)
{
	retrieveArgs retVal;
	for(int i = 2; i < argc; ++i)
	{
		string temp(argv[i]);
		bool missingArgument = false;
		if(temp == string("-h"))
		{
			if(++i < argc)
			{
				retVal.hostfile = argv[i];
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-b"))
		{
			if(++i < argc)
			{
				retVal.bitsperbyte = str2int(argv[i]);
				if(retVal.bitsperbyte == 1)
					retVal.retrieveByte = retrieveByte1;
				else if(retVal.bitsperbyte == 2)
					retVal.retrieveByte = retrieveByte2;
				else if(retVal.bitsperbyte == 4)
					retVal.retrieveByte = retrieveByte4;
				continue;
			}
			missingArgument = true;
		}
		else if(temp == string("-p"))
		{
			if(++i < argc)
			{
				string password = argv[i];
				retVal.encryption = new encryptClass(password);
				continue;
			}
			missingArgument = true;
		}
		if(missingArgument)
			cout << "The option " << temp << " is missing an argument!";
		else
			cout << "Unknown option \"" << temp << "\"!";
		cout << endl;
		retVal.error = true;
		return retVal;
	}
	return retVal;
}

//For 16 bit images:
//This function returns a byte extracted from 6 bytes read from input
unsigned char retrieve16Byte1(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned short readData;
	//Get the first 6 bits
	for(int i = 0; i < 6; i += 3)
	{
		input.read(reinterpret_cast<char*>(&readData), 2);

		retVal |=
			  ((readData & 0x0400) >> (10 - i))
			| ((readData & 0x0020) >> (4 - i)) // == 5 - (i + 1)
			| ((readData & 0x0001) << (i + 2));
	}
	//Get the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);

	if(encryption)
	{
		retVal |= ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2);
		encryption->exe(retVal);
		return retVal;
	}
	else
		return static_cast<unsigned char>(retVal | ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2));
}

//This function returns a byte extracted from 4 bytes read from input
unsigned char retrieve16Byte2(ifstream &input, encryptClass *encryption)
{
	//Get the first 6 bits
	unsigned short readData;
	input.read(reinterpret_cast<char*>(&readData), 2);

	unsigned char retVal = ((readData & 0x0C00) >> 10)
		| ((readData & 0x0060) >> 3)
		| ((readData & 0x0003) << 4);

	//Get the last 2 bits
	input.read(reinterpret_cast<char*>(&readData), 2);

	if(encryption)
	{
		retVal |= ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2);
		encryption->exe(retVal);
		return retVal;
	}
	else
		return static_cast<unsigned char>(retVal | ((readData & 0x0400) >> 4) | ((readData & 0x0020) << 2));
}

//For 24 and 32 bit images:
//This function returns a byte extracted from 8 bytes read from input
unsigned char retrieveByte1(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; ++i)
	{
		//Get value of active bit
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x01) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

//This function returns a byte extracted from 4 bytes read from input
unsigned char retrieveByte2(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; i += 2)
	{
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x03) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

//This function returns a byte extracted from 2 bytes read from input
unsigned char retrieveByte4(ifstream &input, encryptClass *encryption)
{
	unsigned char retVal = 0;
	unsigned char readByte;
	for(int i = 0; i < 8; i += 4)
	{
		input.read(reinterpret_cast<char *>(&readByte), 1);
		retVal |= (readByte & 0x0F) << i;
	}
	if(encryption)
		encryption->exe(retVal);
	return retVal;
}

void showRetrieveHelp()
{
	cout << "Possible options (explanations are in brackets):\n"
		<< "-h hostfile (image file which contains hidden data)\n"
		<< "-p password (password will be used to decrypt the file)\n"
		<< "-b bitsperbyte (defines how many bits of each byte of the hostfile will be used to hide data. Possible values are 1, 2 & 4 (only 1 & 2 for 16 bit images).\nDefault: 1)\n";

	cout << "Non-optional options: -h" << endl;
}
